/*
    XLode, preview release 0.1 - a humble clone of the best game ever.
    Copyright (C) 2001  Hisham Hashem Muhammad
                                                                                
    This program is free software; you can redistribute it and/or modify        
    it under the terms of the GNU General Public License as published by        
    the Free Software Foundation; either version 2 of the License, or           
    (at your option) any later version.                                         
                                                                                
    This program is distributed in the hope that it will be useful,             
    but WITHOUT ANY WARRANTY; without even the implied warranty of              
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               
    GNU General Public License for more details.                                
                                                                                
    You should have received a copy of the GNU General Public License           
    along with this program; if not, write to the Free Software                 
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   
*/

/* ------------------------------------------------------------------------- */
/* Includes */
/* ------------------------------------------------------------------------- */

/* ***** Standard C ***** */

#include <stdlib.h>
#include <stdio.h>

/* ***** Allegro ***** */

#include <allegro.h>

/* ***** JGMod ***** */

#include <jgmod.h>

/* ------------------------------------------------------------------------- */
/* Definitions */
/* ------------------------------------------------------------------------- */

#undef TRUE

#undef FALSE

#ifndef NULL
#define NULL 0
#endif

#define XLode_debug(x) fprintf(stderr, x)
#define XLode_debug1(x,x1) fprintf(stderr, x,x1)
#define XLode_debug2(x,x1,x2) fprintf(stderr, x,x1,x2)
#define XLode_debug3(x,x1,x2,x3) fprintf(stderr, x,x1,x2,x3)
#define XLode_debug4(x,x1,x2,x3,x4) fprintf(stderr, x,x1,x2,x3,x4)

// #define randomNumber(x) ((int)(x * rand()/(RAND_MAX+1.0)))
#define randomNumber(x) (rand() % x)

#define writeText(y,t) textout_centre(pr_backBuffer, font, t, Ci_resX / 2, y, 0xffffff)

/* ***** Game parameters ***** */

#define USE_DOUBLE_BUFFERING

#define USE_BACKGROUND

/**
 * Limitations
 */

#define Ci_resX 640
#define Ci_resY 480

#define Ci_arenaW 40
#define Ci_arenaH 20

#define Ci_spriteW 16
#define Ci_spriteH 20

#define Ci_stageCount 3

#define Ci_figureCount 9

#define Ci_monkCount 10

/**
 * Motion properties of a game element
 */

#define FROM_SIDES  1
#define FROM_UP     2
#define FROM_DOWN   4
#define TO_SIDES    8
#define TO_UP       16
#define TO_DOWN     32
#define FALL_FROM   64
#define FALL_TO     128

/**
 * Amount of time that the player can't move while breaking a brick
 */

#define Ci_breakingTime 5

/**
 * Number of songs used in the game
 */

#define Ci_songCount 5

/* ------------------------------------------------------------------------ */
/* Data structures */
/* ------------------------------------------------------------------------ */

/* ***** Enumerations ***** */

typedef enum {
   LODE = 0,
   MONK,
   BRICK,
   ROCK,
   GHOSTBRICK,
   EMPTY,
   STAIRS,
   TREASURE,
   ROPE
} ElementType;

typedef enum {
   GET_TREASURE = 0,
   BREAK_BRICK,
   REACH_SKY,
   READY,
   GO,
   DEATH,
   YESSS
} SampleType;
#define Ci_sampleCount 7

typedef enum {
   NONE,
   UP,
   DOWN,
   LEFT,
   RIGHT
} Direction;

typedef enum {
   FALSE = 0,
   TRUE
} Boolean;

/* ***** Records ***** */

typedef struct {
   ElementType e_type;
   int i_moveSet;
   int i_status;
} Element;

typedef struct {
   int i_x;
   int i_y;
   int i_offsetX;
   int i_offsetY;
   Direction e_direction;
   Direction e_nextDirection;
   Direction e_lastDirection;
   Boolean b_falling;
   BITMAP** papr_runningLeft;
   BITMAP** papr_runningRight;
   BITMAP* pr_defaultFigure;
   BITMAP* pr_defaultFalling;
   Boolean b_isLode;
} Player;

typedef struct {
   int i_x;
   int i_y;
   int i_timer;
   int i_step;
} BrickEvent;

/* ------------------------------------------------------------------------ */
/* Game data */
/* ------------------------------------------------------------------------ */

/**
 * Generic game data
 */

Element aar_arena[Ci_arenaW][Ci_arenaH];

char aac_mapCache[Ci_stageCount][Ci_arenaW][Ci_arenaH];

Boolean b_retry;

Boolean b_gameOver;

int i_stage;

int i_score;

int i_lives;

FILE* af_stages[Ci_stageCount];
Boolean ab_cachedStage[Ci_stageCount];

Player r_lode;

int i_monkCount;

Player ar_monks[Ci_monkCount];

int i_treasures;

int i_gameTimer;

Boolean b_playing;

Boolean b_victory;

int i_breaking;

Direction e_scheduledBreak;

Direction e_lastBreak;

/**
 * Broken bricks
 */

int i_brokenCount;

int i_brokenHead, i_brokenTail;

BrickEvent ar_broken[Ci_arenaW * Ci_arenaH];

/**
 * Frames for the "breaking bricks" animation
 */

// Concepts:
// The animation contains eight stages, called "steps", but only 5 different
// images are used; hence the diferentiation between the steps of the animation
// and the frames that are actually used.

#define Ci_brokenSteps 8
int ai_brokenTimes[Ci_brokenSteps]  = { 2, 4, 6, 8, 120, 130, 140, 150 };
int ai_brokenFrames[Ci_brokenSteps] = { 1, 2, 3, 0,   3,   2,   1,   4 };

/**
 * Monk logic
 */

Direction ae_directions[4] = {UP, DOWN, LEFT, RIGHT};

/** ***** Graphics ***** */

/**
 * Numbers
 */

BITMAP* apr_numbers[10];

/**
 * Lode animations
 */

BITMAP* apr_lodeRunningRight[8];

BITMAP* apr_lodeRunningLeft[8];

BITMAP* pr_lodeFalling;

/**
 * Monk animations
 */

BITMAP* apr_monkRunningRight[8];

BITMAP* apr_monkRunningLeft[8];

BITMAP* pr_monkFalling;

/**
 * Broken bricks
 */

#define Ci_brokenFrames 5
BITMAP* apr_brokenAnimation[Ci_brokenFrames];

BITMAP* pr_lodeBreakingLeft;
BITMAP* pr_lodeBreakingRight;

/**
 * Generic game graphics
 */

BITMAP* apr_figures[Ci_figureCount];

BITMAP* pr_backBuffer;

BITMAP* pr_arenaBuffer;

/**
 * Graphics for the opening
 */

BITMAP* pr_title;
BITMAP* pr_options;
BITMAP* pr_lodeIcon;

#ifdef USE_BACKGROUND
BITMAP* pr_background;
#endif

BITMAP* pr_gameOver;
BITMAP* pr_gameOverFlip;

/* ***** Sounds ***** */

int i_lodeSoundX;

int i_lodeSoundY;

SAMPLE* pr_lodeVoice;

Boolean b_wasFalling;

JGMOD* apr_songs[Ci_songCount];

SAMPLE* apr_samples[Ci_sampleCount];

/**
 * Lode on the rope
 */

#define Ci_ropeSampleCount 8
SAMPLE* apr_ropeSamples[Ci_ropeSampleCount];

/**
 * Lode falling
 */

#define Ci_fallSampleCount 5
SAMPLE* apr_fallSamples[Ci_fallSampleCount];

/**
 * End of a fall ("ouch")
 */

#define Ci_ouchSampleCount 3
SAMPLE* apr_ouchSamples[Ci_ouchSampleCount];

/**
 * Lode on the stairs
 */

#define Ci_stairsSampleCount 4
SAMPLE* apr_stairsSamples[Ci_stairsSampleCount];

/**
 * Lode running
 */

#define Ci_runningSampleCount 3
SAMPLE* apr_runningSamples[Ci_runningSampleCount];

/* ------------------------------------------------------------------------- */
/* Utility libraries */
/* ------------------------------------------------------------------------- */

/* ***** Allegro-specific utility functions ***** */

/**
 * Configure video to the indicated mode/resolution/depth.
 */

int Allegro_setVideo(int i_depth_) {
   set_color_depth(i_depth_);
   set_gfx_mode(GFX_AUTODETECT, Ci_resX, Ci_resY, 0, 0);
}

/**
 * Initialize all Allegro subsystems and setup the screen.
 */

int Allegro_init() {
   // TODO: check return values
   allegro_init();
   install_keyboard();
   install_timer();
   srand(time(0));
   
   reserve_voices(16, -1);
   install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);
   install_mod(8);
   set_volume(255, -1);
   
   Allegro_setVideo(16);
   return TRUE;
}

/**
 * Finalize Allegro.
 */

int Allegro_done() {
   allegro_exit();
}

/* ------------------------------------------------------------------------- */
/* Game specific routines */
/* ------------------------------------------------------------------------- */

/* ***** Animation ***** */

/*
void Animation_play(char* sz_fileName_, JGMOD* pr_module) {
   BITMAP* pr_centered;

   clear(screen);
   Allegro_setVideo(8);
   pr_centered = create_sub_bitmap(screen, Ci_resX/2-(320/2),Ci_resY/2-(200/2), 320, 200);
   set_mod_volume(250);
   play_mod(pr_module, FALSE);
   play_fli(sz_fileName_, pr_centered, 0, NULL);
   destroy_bitmap(pr_centered);
   Allegro_setVideo(16);
}
*/

void Animation_play(char* sz_fileName_, JGMOD* pr_module) {
   BITMAP* pr_centered;

   set_color_conversion(COLORCONV_EXPAND_256);

   clear(screen);
   pr_centered = create_sub_bitmap(screen, Ci_resX/2-(320/2),Ci_resY/2-(200/2), 320, 200);
   set_mod_volume(250);
   play_mod(pr_module, FALSE);
   open_fli(sz_fileName_);
   while (next_fli_frame(0) == FLI_OK) {
      set_palette(fli_palette);
      rest(30);
      blit(fli_bitmap, pr_centered, 0, 0, 0, 0, 320, 200);
   }
   close_fli();
   destroy_bitmap(pr_centered);
}

/* ***** Figures ***** */

/**
 * Draws the figure that's at the indicated position
 * directly at the back buffer.
 */

void Figure_draw(int i_x_, int i_y_) {
   Element* pr_this = &aar_arena[i_x_][i_y_];
#ifdef USE_BACKGROUND
   blit(pr_background, pr_backBuffer, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, Ci_spriteW, Ci_spriteH);
#else
   rectfill(pr_backBuffer,
            i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, 
            i_x_ * Ci_spriteW + Ci_spriteW - 1, i_y_ * Ci_spriteH + Ci_spriteH - 1, 
            0x000000);
#endif
   if (apr_figures[pr_this->e_type] != NULL) {
      draw_sprite(pr_backBuffer, apr_figures[pr_this->e_type], i_x_ * Ci_spriteW, i_y_ * Ci_spriteH);
   }
}

/**
 * Draws the figure that's at the indicated position,
 * obtaining it from the arena's cache.
 */

void Figure_show(int i_x_, int i_y_) {
   blit(pr_arenaBuffer, pr_backBuffer, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, Ci_spriteW, Ci_spriteH);
}

/**
 * Updates the arena's cache putting the given sprite at
 * the specified position.
 */

void Figure_update(int i_x_, int i_y_, BITMAP* pr_sprite_) {
   Element* pr_this = &aar_arena[i_x_][i_y_];
#ifdef USE_BACKGROUND
   blit(pr_background, pr_arenaBuffer, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, Ci_spriteW, Ci_spriteH);
#else
   rectfill(pr_arenaBuffer,
            i_x_ * Ci_spriteW, i_y_ * Ci_spriteH, 
            i_x_ * Ci_spriteW + Ci_spriteW - 1, i_y_ * Ci_spriteH + Ci_spriteH,
            0x000000);
#endif
   if (pr_sprite_ != NULL) {
      draw_sprite(pr_arenaBuffer, pr_sprite_, i_x_ * Ci_spriteW, i_y_ * Ci_spriteH);
   }
}

void Number_displayAt(BITMAP* pr_target, int i_x_, int i_y_, int i_number_, int i_digits_) {
   char sz_written[20];
   int i_length;
   int i_pad;
   int i;

   i_pad = 0;   
   sprintf(sz_written, "%d", i_number_);
   i_length = strlen(sz_written);
   if (i_length < i_digits_) {
      for(; i_pad < i_digits_ - i_length; i_pad++) {
         draw_sprite(pr_target, apr_numbers[0], i_x_+(i_pad*13), i_y_);
      }
   }
   for(i = 0; i < i_length; i++) {
      draw_sprite(pr_target, apr_numbers[sz_written[i]-48], i_x_+(i_pad*13), i_y_);
      i_pad++;
   }
}

/* ***** Sounds ***** */

void Sound_playSample(SampleType e_id_, int i_xPosition_) {
   play_sample(apr_samples[e_id_], 255, 255 * ((double)i_xPosition_ / (double)Ci_arenaW), 1000, FALSE);
}

/**
 * Play a sound base on Lode's current position
 */

void Sound_lodeSound() {
   int i_height;
   int i_yCount;
   int i_fall;
   int i_panning;
   SAMPLE* pr_newLodeVoice;

   pr_newLodeVoice = NULL;
   
   i_panning = 255 * ((double)r_lode.i_x / (double)Ci_arenaW);

   if (i_lodeSoundX == r_lode.i_x && i_lodeSoundY == r_lode.i_y)
      return;

   i_lodeSoundX = r_lode.i_x;
   i_lodeSoundY = r_lode.i_y;

   if (aar_arena[r_lode.i_x][r_lode.i_y].e_type == ROPE) {
      pr_newLodeVoice = apr_ropeSamples[randomNumber(Ci_ropeSampleCount)];
   } else if (aar_arena[r_lode.i_x][r_lode.i_y].e_type == STAIRS) {
      pr_newLodeVoice = apr_stairsSamples[randomNumber(Ci_stairsSampleCount)];
   } else if (r_lode.b_falling && b_wasFalling) {
      // Check if done falling
      i_yCount = r_lode.i_y + 1;
      // TODO: this entire function sucks at 04:16am
      if (!(i_yCount <= Ci_arenaH && aar_arena[r_lode.i_x][i_yCount].i_moveSet & FALL_TO
          && i_yCount <= Ci_arenaH && aar_arena[r_lode.i_x][i_yCount].i_moveSet & FALL_FROM)) {
         pr_newLodeVoice = apr_ouchSamples[randomNumber(Ci_ouchSampleCount)];
         b_wasFalling = FALSE;
      }
   } else if (r_lode.b_falling && !b_wasFalling) {
      // Scream
      i_height = 0;
      i_yCount = r_lode.i_y;
      // TODO: fix this; it's 3:42am and I don't want to think about it
      while (i_yCount <= Ci_arenaH && aar_arena[r_lode.i_x][i_yCount].i_moveSet & FALL_TO
          && i_yCount <= Ci_arenaH && aar_arena[r_lode.i_x][i_yCount].i_moveSet & FALL_FROM) {
         i_height++;
         i_yCount++;
      }
      if (i_height <= 2) i_fall = 0;
      else if (i_height <= 5) i_fall = 1;
      else if (i_height <= 8) i_fall = 2;
      else if (i_height <= 11) i_fall = 3;
      else i_fall = 4;
      pr_newLodeVoice = apr_fallSamples[i_fall];
      b_wasFalling = TRUE;
   } else {
      pr_newLodeVoice = apr_runningSamples[randomNumber(Ci_runningSampleCount)];
   }
   if (pr_newLodeVoice != NULL) {
      stop_sample(pr_lodeVoice);
      play_sample(pr_newLodeVoice, 255, i_panning, 1000, 0);
      pr_lodeVoice = pr_newLodeVoice;
   }
}

/* ***** Bricks ***** */

/**
 * Break a brick: make the space empty and schedule the bricks reappearance.
 */

void Bricks_break(int i_x_, int i_y_) {
   BrickEvent* pr_brick = &(ar_broken[i_brokenTail]);

   Sound_playSample(BREAK_BRICK, i_x_);
   
   i_brokenCount++;
   pr_brick->i_x = i_x_;
   pr_brick->i_y = i_y_;
   pr_brick->i_timer = i_gameTimer;
   pr_brick->i_step = 0;
   i_brokenTail++;
   aar_arena[i_x_][i_y_].e_type = EMPTY;
   aar_arena[i_x_][i_y_].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                   | TO_SIDES             | TO_DOWN
                                   | FALL_FROM  | FALL_TO;
}

INLINE void Bricks_update() {
   int i;
   int i_x, i_y;
      
   if (i_brokenCount == 0)
      return;

   for (i = i_brokenHead; i < i_brokenTail; i++) {
      if (ar_broken[i].i_timer + ai_brokenTimes[ar_broken[i].i_step] == i_gameTimer) {
         Figure_update(ar_broken[i].i_x,
                       ar_broken[i].i_y,
                       apr_brokenAnimation[ai_brokenFrames[ar_broken[i].i_step]]);
         Figure_show(ar_broken[i].i_x, ar_broken[i].i_y);
         ar_broken[i].i_step++;
      }
   }

   if (ar_broken[i_brokenHead].i_step == Ci_brokenSteps) {
      i_x = ar_broken[i_brokenHead].i_x;
      i_y = ar_broken[i_brokenHead].i_y;
      aar_arena[i_x][i_y].e_type = BRICK;
      aar_arena[i_x][i_y].i_moveSet = 0;
      i_brokenHead++;
      i_brokenCount--;
      if (i_brokenCount == 0) {
         i_brokenHead = 0;
         i_brokenTail = 0;
      }
      // TODO: check death of lode or monk
   }

}

/* ***** Player ***** */

/**
 * Verifies that the attempted move is valid in the arena.
 */

int Player_isValidMove(Player r_this, Direction e_direction) {

   int i_x, i_lx;
   int i_y, i_ly;
   
   if (e_direction == NONE)
      return TRUE;
   
   i_x = r_this.i_x;
   i_lx = i_x;
   i_y = r_this.i_y;
   i_ly = i_y;
   if (e_direction == UP) i_y--;
   if (e_direction == DOWN) i_y++;
   if (e_direction == LEFT) i_x--;
   if (e_direction == RIGHT) i_x++;
   
   if (i_x < 0 || i_x > Ci_arenaW - 1 || i_y < 0 || i_y > Ci_arenaH - 1)
      return FALSE;

   if (e_direction == LEFT || e_direction == RIGHT)
      return (aar_arena[i_lx][i_ly].i_moveSet & TO_SIDES
              && aar_arena[i_x][i_y].i_moveSet & FROM_SIDES);
   if (e_direction == UP)
      return (aar_arena[i_lx][i_ly].i_moveSet & TO_UP
              && aar_arena[i_x][i_y].i_moveSet & FROM_DOWN);
   if (e_direction == DOWN)
      return (aar_arena[i_lx][i_ly].i_moveSet & TO_DOWN
              && aar_arena[i_x][i_y].i_moveSet & FROM_UP);
}

/**
 * Erase the area surrounding the player, restoring
 * the arena. Notice that other characters (monks)
 * will be erased too.
 */

void Player_clear(Player r_this) {
   Figure_show(r_this.i_x, r_this.i_y);
   if (r_this.i_offsetX > 0)
      Figure_show(r_this.i_x+1, r_this.i_y);
   else if (r_this.i_offsetX < 0)
      Figure_show(r_this.i_x-1, r_this.i_y);
   else if (r_this.i_offsetY > 0)
      Figure_show(r_this.i_x, r_this.i_y+1);
   else if (r_this.i_offsetY < 0)
      Figure_show(r_this.i_x, r_this.i_y-1);
}

/**
 * Draw the player.
 */

void Player_draw(Player r_this) {
   int i_index;
   int i_rope;
   BITMAP* pr_sprite;
   BITMAP** apr_runningLeft;
   BITMAP** apr_runningRight;

   apr_runningLeft = r_this.papr_runningLeft;
   apr_runningRight = r_this.papr_runningRight;
   
   // Remember that only one of those will be > 0
   i_index = 
   ABS( r_this.i_offsetX / 4 + r_this.i_offsetY / 5 );
   
   if (aar_arena[r_this.i_x][r_this.i_y].e_type == ROPE)
      i_rope = 4;
   else
      i_rope = 0;

   // TODO: stairs
   if (r_this.e_lastDirection == LEFT)
      pr_sprite = apr_runningLeft[i_index + i_rope];
   else if (r_this.e_lastDirection == RIGHT)
      pr_sprite = apr_runningRight[i_index + i_rope];
   else if (r_this.e_lastDirection == UP)
      pr_sprite = r_this.pr_defaultFalling;
   else if (r_this.e_lastDirection == DOWN) {
      if (aar_arena[r_this.i_x][r_this.i_y].i_moveSet & FALL_FROM)
         pr_sprite = r_this.pr_defaultFalling;
      else
         pr_sprite = r_this.pr_defaultFalling;
   }
   
   if (r_this.b_isLode && i_breaking > 0) {
      if (e_lastBreak == LEFT)
         pr_sprite = pr_lodeBreakingLeft;
      else
         pr_sprite = pr_lodeBreakingRight;
   }
   
   draw_sprite(pr_backBuffer,
               pr_sprite,
               Ci_spriteW * r_this.i_x + r_this.i_offsetX,
               Ci_spriteH * r_this.i_y + r_this.i_offsetY);
}

/**
 * Updates the player's position, based on current
 * direction coordinates.
 */

// Concepts:
// Stable - player is exactly in a position in the arena, with no offset.
// Unstable - player is moving between two positions, as indicated by the 
// offset.

void Player_move(Player* pr_this) {
   // If in a stable place, check if direction is valid and update it.
   if (pr_this->i_offsetX == 0 && pr_this->i_offsetY == 0) {
      if (pr_this->i_y < Ci_arenaH
          && aar_arena[pr_this->i_x][pr_this->i_y].i_moveSet & FALL_FROM
          && aar_arena[pr_this->i_x][pr_this->i_y+1].i_moveSet & FALL_TO
         ) {
         // Gravity has the priority :)
         pr_this->e_direction = DOWN;
         pr_this->e_nextDirection = DOWN;
         pr_this->b_falling = TRUE;
      } else {
         if (pr_this->b_falling) {
            pr_this->e_direction = NONE;
            pr_this->e_nextDirection = NONE;
         } else if (Player_isValidMove(*pr_this, pr_this->e_nextDirection)) {
            pr_this->e_direction = pr_this->e_nextDirection;
         } else {
            pr_this->e_direction = NONE;
            pr_this->e_nextDirection = NONE;
         }
         pr_this->b_falling = FALSE;
      }
   } else {
      // If opposite direction, allow changing direction even if unstable
      if (pr_this->e_direction == UP && pr_this->e_nextDirection == DOWN)
         pr_this->e_direction = DOWN;
      if (!pr_this->b_falling && pr_this->e_direction == DOWN && pr_this->e_nextDirection == UP)
         pr_this->e_direction = UP;
      if (pr_this->e_direction == LEFT && pr_this->e_nextDirection == RIGHT)
         pr_this->e_direction = RIGHT;
      if (pr_this->e_direction == RIGHT && pr_this->e_nextDirection == LEFT)
         pr_this->e_direction = LEFT;
   }
   // Move
   if (pr_this->e_direction == RIGHT)
      pr_this->i_offsetX += 4;
   if (pr_this->e_direction == LEFT)
      pr_this->i_offsetX -= 4;
   if (pr_this->e_direction == DOWN)
      pr_this->i_offsetY += 5;
   if (pr_this->e_direction == UP)
      pr_this->i_offsetY -= 5;
   // If full offset is reached, stabilize position
   if (pr_this->i_offsetX == Ci_spriteW) {
      pr_this->i_x++;
      pr_this->i_offsetX = 0;
   }
   if (pr_this->i_offsetX == -Ci_spriteW) {
      pr_this->i_x--;
      pr_this->i_offsetX = 0;
   }
   if (pr_this->i_offsetY == Ci_spriteH) {
      pr_this->i_y++;
      pr_this->i_offsetY = 0;
   }
   if (pr_this->i_offsetY == -Ci_spriteH) {
      pr_this->i_y--;
      pr_this->i_offsetY = 0;
   }
   if (pr_this->e_direction != NONE)
      pr_this->e_lastDirection = pr_this->e_direction;
}

/**
 * Returns TRUE if the player is perfectly centered in a position.
 */

INLINE int Player_isStable(Player r_this) {
   return (r_this.i_offsetX == 0 && r_this.i_offsetY == 0);
}

/**
 * Blink all players on the screen.
 */

INLINE void Player_blinkAll(int i_times_, int i_onTime_, int i_offTime_) {
   int i;
   int j;
   for (j = 0; j < i_times_; j++) {
      Player_clear(r_lode);
      for (i = 0; i < i_monkCount; i++)
         Player_clear(ar_monks[i]);
#ifdef USE_DOUBLE_BUFFERING
      blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);
#endif
      rest(i_offTime_);
      Player_draw(r_lode);
      for (i = 0; i < i_monkCount; i++)
         Player_draw(ar_monks[i]);
#ifdef USE_DOUBLE_BUFFERING
      blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);
#endif
      rest(i_onTime_);
   }
}

/* ***** Lode specific functions ***** */

/**
 * Attempt to break a brick in a specified direction, which may be
 * LEFT or RIGHT.
 */

void Lode_break(Direction e_direction) {
   int i_x;
   int i_y;

   i_x = r_lode.i_x;
   i_y = r_lode.i_y;

   if (i_y == Ci_arenaH)
      return;
   if (e_direction == LEFT) {
      if (i_x == 0)
         return;
      if (aar_arena[i_x-1][i_y].i_moveSet & TO_SIDES
         && aar_arena[i_x-1][i_y+1].e_type == BRICK
      ) {
         Bricks_break(i_x-1, i_y+1);
      }
   } else if (e_direction == RIGHT) {
      if (i_x == Ci_arenaW - 1)
         return;
      if (aar_arena[i_x+1][i_y].i_moveSet & TO_SIDES
         && aar_arena[i_x+1][i_y+1].e_type == BRICK
      ) {
         Bricks_break(i_x+1, i_y+1);
      }
   }
}

/* ***** Monk specific functions ***** */

Direction Monk_decideDirection(Player r_monk) {
   int i_xDistance;
   int i_yDistance;
   Direction e_choice;
   Boolean ab_isValid[4];
   int i;
   int i_limit;
   
   i_xDistance = r_monk.i_x - r_lode.i_x;
   i_yDistance = r_monk.i_y - r_lode.i_y;
   
   for(i=0; i<4; i++)
      ab_isValid[i] = Player_isValidMove(r_monk, ae_directions[i]);
   
   // Case 1: Lode is in sight
   if (ab_isValid[0] && i_xDistance == 0)
      return UP;
   if (ab_isValid[1] && i_xDistance == 0)
      return DOWN;
   if (ab_isValid[2] && i_yDistance == 0)
      return LEFT;
   if (ab_isValid[3] && i_yDistance == 0)
      return RIGHT;
   // Case 2: Know when Lode is close
   if (ab_isValid[2] && i_xDistance < 5 && ABS(i_yDistance) < 5)
      return RIGHT;
   if (ab_isValid[3] && i_yDistance > 5 && ABS(i_yDistance) < 5)
      return LEFT;
   if (ab_isValid[0] && i_yDistance > 0)
      if (randomNumber(2))
         return UP;
   if (ab_isValid[1] && i_yDistance < 0)
      if (randomNumber(2))
         return DOWN;
   i_limit = 0;
   do {
      if (randomNumber(2))
         e_choice = ae_directions[randomNumber(4)];
      else
         e_choice = r_monk.e_lastDirection;
      i_limit++;
   } while (i_limit < 16 && !Player_isValidMove(r_monk, e_choice));
   if (i_limit == 16) e_choice = NONE;
   
   return e_choice;
}

/* ***** Treasure ***** */

INLINE void Treasure_check() {
   if (aar_arena[r_lode.i_x][r_lode.i_y].e_type == TREASURE) {
      i_treasures--;
      if (i_treasures == 0) {
         Sound_playSample(REACH_SKY, r_lode.i_x);
      } else {
         Sound_playSample(GET_TREASURE, r_lode.i_x);
      }
      i_score+=50;
      aar_arena[r_lode.i_x][r_lode.i_y].e_type = EMPTY;
      Figure_update(r_lode.i_x, r_lode.i_y,
                    apr_figures[EMPTY]);
      Figure_show(r_lode.i_x, r_lode.i_y);
   }   
}   

/* ***** Status bar ***** */

void Status_displayScore() {
   blit(pr_arenaBuffer, pr_backBuffer, 165, 428, 165, 428, 8*14, 20);
   Number_displayAt(pr_backBuffer, 165, 428, i_score, 7);
}

/* ***** Gameplay ***** */

/**
 * Prepare for the start of a new round of gameplay:
 * clear everything and start at the first stage.
 */

void Play_init() {
   b_retry = FALSE;
   b_gameOver = FALSE;
   i_stage = 1;
   i_score = 0;
   i_lives = 3;
}

/**
 * Load stage file; construct the arena.
 */

void Play_loadStage() {
   char sz_fileName[20];
   int i, j;
   char c_mapItem;
   
   i_brokenCount = 0;
   i_brokenHead = 0;
   i_brokenTail = 0;
   i_treasures = 0;

   sprintf(sz_fileName, "stage%d.txt", i_stage);
   
   clear(screen);
   
   clear(pr_backBuffer);
#ifdef USE_BACKGROUND
   blit(pr_background, pr_backBuffer, 0, 0, 0, 0, Ci_resX, Ci_resY);
#endif
   
   i_monkCount = 0;

   for(j = 0; j < Ci_arenaH; j++) {
      for(i = 0; i < Ci_arenaW; i++) {
         if (ab_cachedStage[i_stage]) {
            c_mapItem = aac_mapCache[i_stage][i][j];
         } else {
            do {
               if (!feof(af_stages[i_stage])) {
                  fread(&c_mapItem, sizeof(char), 1, af_stages[i_stage]);
               } else
                  c_mapItem = ' '; // Incomplete file!?
            } while (c_mapItem == '\n');
            aac_mapCache[i_stage][i][j] = c_mapItem;
         }
         switch (c_mapItem) {
         case 'L':
            aar_arena[i][j].e_type = EMPTY;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            r_lode.i_x = i;
            i_lodeSoundX = r_lode.i_x;
            r_lode.i_y = j;
            i_lodeSoundY = r_lode.i_y;
            r_lode.i_offsetX = 0;
            r_lode.i_offsetY = 0;
            r_lode.b_falling = FALSE;
            b_wasFalling = FALSE;
            r_lode.e_direction = NONE;
            r_lode.e_nextDirection = NONE;
            r_lode.e_lastDirection = RIGHT;
            r_lode.papr_runningLeft = apr_lodeRunningLeft;
            r_lode.papr_runningRight = apr_lodeRunningRight;
            r_lode.pr_defaultFigure = apr_figures[LODE];
            r_lode.pr_defaultFalling = pr_lodeFalling;
            r_lode.b_isLode = TRUE;
            break;
         case 'M':
            aar_arena[i][j].e_type = EMPTY;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            ar_monks[i_monkCount].i_x = i;
            ar_monks[i_monkCount].i_y = j;
            ar_monks[i_monkCount].i_offsetX = 0;
            ar_monks[i_monkCount].i_offsetY = 0;
            ar_monks[i_monkCount].b_falling = FALSE;
            ar_monks[i_monkCount].e_direction = NONE;
            ar_monks[i_monkCount].e_nextDirection = NONE;
            ar_monks[i_monkCount].e_lastDirection = LEFT;
            ar_monks[i_monkCount].papr_runningLeft = apr_monkRunningLeft;
            ar_monks[i_monkCount].papr_runningRight = apr_monkRunningRight;
            ar_monks[i_monkCount].pr_defaultFigure = apr_figures[MONK];
            ar_monks[i_monkCount].pr_defaultFalling = pr_monkFalling;
            ar_monks[i_monkCount].b_isLode = FALSE;
            i_monkCount++;
            break;
         case 'B':
            aar_arena[i][j].e_type = BRICK;
            aar_arena[i][j].i_moveSet = 0;
            break;
         case 'R':
            aar_arena[i][j].e_type = ROCK;
            aar_arena[i][j].i_moveSet = 0;
            break;
         case 'G':
            aar_arena[i][j].e_type = GHOSTBRICK;
            aar_arena[i][j].i_moveSet =              FROM_UP | FROM_DOWN
                                                             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            break;
         case ' ':
            aar_arena[i][j].e_type = EMPTY;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            break;
         case 'S':
            aar_arena[i][j].e_type = STAIRS;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES   | TO_UP   | TO_DOWN;
            break;
         case 'T':
            aar_arena[i][j].e_type = TREASURE;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            i_treasures++;
            break;
         case '-':
            aar_arena[i][j].e_type = ROPE;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                                   | FALL_TO;
            break;
         default:
            // Invalid file!?
            aar_arena[i][j].e_type = EMPTY;
            aar_arena[i][j].i_moveSet = FROM_SIDES | FROM_UP | FROM_DOWN
                                      | TO_SIDES             | TO_DOWN
                                      | FALL_FROM  | FALL_TO;
            break;
         }
         Figure_draw(i, j);
      }
   }


   blit(pr_backBuffer, pr_arenaBuffer, 0, 0, 0, 0, Ci_resX, Ci_resY);

   Number_displayAt(pr_backBuffer, 165, 428, i_score, 7);

   Number_displayAt(pr_backBuffer, 505, 428, i_stage, 2);

   for (i = 0; i < MIN(i_lives, 6); i++) {
      draw_sprite(pr_backBuffer, apr_lodeRunningRight[0], 323 + (i*18), 428);
   }
   
   blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);

#ifndef USE_DOUBLE_BUFFERING   
   pr_backBuffer = screen;
#endif

   if (!ab_cachedStage[i_stage]) {
      fclose(af_stages[i_stage]);
      ab_cachedStage[i_stage] = TRUE;
   }
   
}

/**
 * This is the actual gameplay loop.
 */

void Play_timedLoop() {
   int i;

   if (key[KEY_ESC])
      b_playing = FALSE;
      
   if (i_breaking > 0)
      i_breaking--;
   else {
      if (key[KEY_UP] || key[KEY_I])
         r_lode.e_nextDirection = UP;
      if (key[KEY_DOWN] || key[KEY_K])
         r_lode.e_nextDirection = DOWN;
      if (key[KEY_LEFT] || key[KEY_J])
         r_lode.e_nextDirection = LEFT;
      if (key[KEY_RIGHT] || key[KEY_L])
         r_lode.e_nextDirection = RIGHT;
      if (key[KEY_U])
         e_scheduledBreak = LEFT;
      if (key[KEY_O])
         e_scheduledBreak = RIGHT;
   }

//   clear_keybuf();

   Player_clear(r_lode);
   for (i = 0; i < i_monkCount; i++)
      Player_clear(ar_monks[i]);

   Player_move(&r_lode);
   
   Bricks_update();

   for (i = 0; i < i_monkCount; i++) {
      if (Player_isStable(ar_monks[i])) {
         ar_monks[i].e_nextDirection = Monk_decideDirection(ar_monks[i]);
      }
      if (ar_monks[i].b_falling || i_gameTimer % 3)
         Player_move(&ar_monks[i]);
      if (
            (ABS((ar_monks[i].i_x * Ci_spriteW + ar_monks[i].i_offsetX) - (r_lode.i_x * Ci_spriteW + r_lode.i_offsetX)) < 8)
         && (ABS((ar_monks[i].i_y * Ci_spriteH + ar_monks[i].i_offsetY) - (r_lode.i_y * Ci_spriteH + r_lode.i_offsetY)) < 10)
          ) {
         b_playing = FALSE;
      }
   }

   if (Player_isStable(r_lode)) {
      Sound_lodeSound();
      Treasure_check();
      if (i_treasures == 0 && r_lode.i_y == 0) {
         b_victory = TRUE;
         b_playing = FALSE;
      }
      if (e_scheduledBreak != NONE) {
         r_lode.e_nextDirection = NONE;
         Lode_break(e_scheduledBreak);
         i_breaking = Ci_breakingTime;
         e_lastBreak = e_scheduledBreak;
         e_scheduledBreak = NONE;
      }
   }

   Player_draw(r_lode);
   for (i = 0; i < i_monkCount; i++)
      Player_draw(ar_monks[i]);

#ifdef USE_DOUBLE_BUFFERING
   blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);
#endif

}

/**
 * Loop for libraries that require polling.
 */

void Play_pollLoop() {
   
   poll_keyboard();

}

/**
 * Inner loop: this is the main gameplay loop.
 * It controls other loops.
 */

int Play_loop() {
   int i;
   int i_savedScore;

   b_playing = TRUE;
   b_victory = FALSE;
   i_gameTimer = 0;
   i_breaking = 0;
   e_scheduledBreak = NONE;

/*
   install_int(Play_timedLoop, 80);
   install_int(Play_pollLoop, 80);

   while (b_playing)
      rest(500);

   remove_int(Play_pollLoop);
   remove_int(Play_timedLoop);
*/

   set_mod_volume(128);
   play_mod(apr_songs[2], TRUE);

   Sound_playSample(READY, 128);
   
   Player_blinkAll(3,300,300);

   Sound_playSample(GO, 128);
   rest(500);

   while (b_playing) {
      i_savedScore = i_score;
      Play_timedLoop();
      if (i_score != i_savedScore) {
         Status_displayScore();
      }
      i_gameTimer++;
      for (i = 0; i < 5; i++) {
      Play_pollLoop();
         rest(10);
      }
   }
   
   stop_sample(pr_lodeVoice);
   stop_mod();
   clear_keybuf();
   
   if (b_victory) {
      play_mod(apr_songs[4], FALSE);
      rest(5000);
      Sound_playSample(YESSS, r_lode.i_x);
      rest(1500);
      return TRUE;
   }
   
   Sound_playSample(DEATH, r_lode.i_x);
   rest(1500);
   return FALSE;
}

void Play_nextStage() {
   i_stage++;
   i_lives++;
}

/* ***** Game functions ***** */

// Debug wrapper for load_bitmap
BITMAP* xload_bitmap(char* filename, PALLETE p) {
   BITMAP* result;
   XLode_debug1("Loading %s... ", filename);
   result = load_bitmap(filename, p);
   XLode_debug(" OK.\n");
   return result;
}

void Game_openingCredits() {
   // Initial credits
   clear(pr_backBuffer);
   writeText(40,"XLODE");
   writeText(120,"(C) 2001 Pixellation Studios");
   writeText(140,"Written by Hisham H. Muhammad");
   writeText(200,"This is free software");
   writeText(220,"released under the terms of the GNU General Public License");
   writeText(240,"as published by the Free Software Foundation;");
   writeText(260,"either version 2 or (at your option) any later version.");
   writeText(280,"This program is released in the hope that it will be");
   writeText(300,"useful, but without any warranty, without even implied");
   writeText(320,"warranty of merchantability or fitness for a particular purpose.");
   writeText(340,"See the GNU General Public License for more details.");
   blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);
}

/**
 * Game initializations; allocates the game data and
 * initializes the utility libraries.
 */

int Game_init() {
   PALLETE r_pallete;
   char sz_fileName[20];
   int i;

   if (! Allegro_init() ) return FALSE;
   
   pr_backBuffer = create_bitmap(Ci_resX, Ci_resY);
   pr_arenaBuffer = create_bitmap(Ci_resX, Ci_resY);
   clear(pr_arenaBuffer);
   clear(pr_backBuffer);
   
   Game_openingCredits();

   pr_title = load_bitmap("title.tga", r_pallete);
   pr_options = load_bitmap("options.tga", r_pallete);
   pr_lodeIcon = load_bitmap("lodeicon.tga", r_pallete);
#ifdef USE_BACKGROUND
   pr_background = load_bitmap("background.tga", r_pallete);
#endif
   pr_gameOver = load_bitmap("gameover.tga", r_pallete);
   pr_gameOverFlip = create_bitmap(240, 40);
   draw_sprite_v_flip(pr_gameOverFlip, pr_gameOver, 0, 0);

   apr_figures[LODE] = load_bitmap("lode.tga", r_pallete);
   apr_figures[MONK] = load_bitmap("monk.tga", r_pallete);
   apr_figures[BRICK] = load_bitmap("brick.tga", r_pallete);
   apr_figures[ROCK] = load_bitmap("rock.tga", r_pallete);
   apr_figures[GHOSTBRICK] = load_bitmap("brick.tga", r_pallete);
   apr_figures[STAIRS] = load_bitmap("stairs.tga", r_pallete);
   apr_figures[TREASURE] = load_bitmap("treasure.tga", r_pallete);
   apr_figures[ROPE] = load_bitmap("rope.tga", r_pallete);
   apr_figures[EMPTY] = NULL;
   
   apr_samples[GET_TREASURE] = load_sample("treasure.wav");
   apr_samples[BREAK_BRICK] = load_sample("dig.wav");
   apr_samples[REACH_SKY] = load_sample("reach.wav");
   apr_samples[READY] = load_sample("ready.wav");
   apr_samples[GO] = load_sample("go.wav");
   apr_samples[DEATH] = load_sample("death.wav");
   apr_samples[YESSS] = load_sample("yesss.wav");
   apr_songs[0] = load_mod("intro.xm");
   apr_songs[1] = load_mod("song.xm");
   apr_songs[2] = load_mod("groove.xm");
   apr_songs[3] = load_mod("gameover.xm");
   apr_songs[4] = load_mod("victory.xm");

   apr_brokenAnimation[0] = NULL;
   apr_brokenAnimation[1] = load_bitmap("brick1.tga", r_pallete);
   apr_brokenAnimation[2] = load_bitmap("brick2.tga", r_pallete);
   apr_brokenAnimation[3] = load_bitmap("brick3.tga", r_pallete);
   apr_brokenAnimation[4] = load_bitmap("brick.tga", r_pallete);

   pr_lodeBreakingLeft = load_bitmap("breakl.tga", r_pallete);
   pr_lodeBreakingRight = load_bitmap("breakr.tga", r_pallete);

   // Lode running to the right
   apr_lodeRunningRight[0] = load_bitmap("lode1r.tga", r_pallete);
   apr_lodeRunningRight[1] = load_bitmap("lode2r.tga", r_pallete);
   apr_lodeRunningRight[2] = load_bitmap("lode3r.tga", r_pallete);
   apr_lodeRunningRight[3] = load_bitmap("lode4r.tga", r_pallete);

   // Lode on the rope to the right
   apr_lodeRunningRight[4] = load_bitmap("rlode1r.tga", r_pallete);
   apr_lodeRunningRight[5] = load_bitmap("rlode2r.tga", r_pallete);
   apr_lodeRunningRight[6] = load_bitmap("rlode3r.tga", r_pallete);
   apr_lodeRunningRight[7] = load_bitmap("rlode4r.tga", r_pallete);

   // Lode running to the left
   apr_lodeRunningLeft[0] = load_bitmap("lode1l.tga", r_pallete);
   apr_lodeRunningLeft[1] = load_bitmap("lode2l.tga", r_pallete);
   apr_lodeRunningLeft[2] = load_bitmap("lode3l.tga", r_pallete);
   apr_lodeRunningLeft[3] = load_bitmap("lode4l.tga", r_pallete);

   // Lode on the rope to the left
   apr_lodeRunningLeft[4] = load_bitmap("rlode1l.tga", r_pallete);
   apr_lodeRunningLeft[5] = load_bitmap("rlode2l.tga", r_pallete);
   apr_lodeRunningLeft[6] = load_bitmap("rlode3l.tga", r_pallete);
   apr_lodeRunningLeft[7] = load_bitmap("rlode4l.tga", r_pallete);

   pr_lodeFalling = load_bitmap("lodef.tga", r_pallete);

   // Monk running to the right
   apr_monkRunningRight[0] = load_bitmap("monk1r.tga", r_pallete);
   apr_monkRunningRight[1] = load_bitmap("monk2r.tga", r_pallete);
   apr_monkRunningRight[2] = load_bitmap("monk3r.tga", r_pallete);
   apr_monkRunningRight[3] = load_bitmap("monk4r.tga", r_pallete);
   
   // Monk on the rope to the right
   apr_monkRunningRight[4] = load_bitmap("rmonk1r.tga", r_pallete);
   apr_monkRunningRight[5] = load_bitmap("rmonk2r.tga", r_pallete);
   apr_monkRunningRight[6] = load_bitmap("rmonk3r.tga", r_pallete);
   apr_monkRunningRight[7] = load_bitmap("rmonk4r.tga", r_pallete);

   // Monk running to the left
   apr_monkRunningLeft[0] = load_bitmap("monk1l.tga", r_pallete);
   apr_monkRunningLeft[1] = load_bitmap("monk2l.tga", r_pallete);
   apr_monkRunningLeft[2] = load_bitmap("monk3l.tga", r_pallete);
   apr_monkRunningLeft[3] = load_bitmap("monk4l.tga", r_pallete);

   // Monk on the rope to the left
   apr_monkRunningLeft[4] = load_bitmap("rmonk1l.tga", r_pallete);
   apr_monkRunningLeft[5] = load_bitmap("rmonk2l.tga", r_pallete);
   apr_monkRunningLeft[6] = load_bitmap("rmonk3l.tga", r_pallete);
   apr_monkRunningLeft[7] = load_bitmap("rmonk4l.tga", r_pallete);

   pr_monkFalling = load_bitmap("monkf.tga", r_pallete);
   
   // Numbers
   for (i = 0; i < 10; i++) {
      sprintf(sz_fileName, "number%d.tga", i);
      apr_numbers[i] = load_bitmap(sz_fileName, r_pallete);
   }
   // Sounds: Lode on the rope
   for (i = 0; i < Ci_ropeSampleCount; i++) {
      sprintf(sz_fileName, "rope%d.wav", i+1);
      apr_ropeSamples[i] = load_sample(sz_fileName);
   }
   // Lode on the stairs
   for (i = 0; i < Ci_stairsSampleCount; i++) {
      sprintf(sz_fileName, "stairs%d.wav", i+1);
      apr_stairsSamples[i] = load_sample(sz_fileName);
   }
   // Lode falling
   for (i = 0; i < Ci_fallSampleCount; i++) {
      sprintf(sz_fileName, "fall%d.wav", i+1);
      apr_fallSamples[i] = load_sample(sz_fileName);
   }
   // End of a fall
   for (i = 0; i < Ci_ouchSampleCount; i++) {
      sprintf(sz_fileName, "ouch%d.wav", i+1);
      apr_ouchSamples[i] = load_sample(sz_fileName);
   }
   // Lode running
   for (i = 0; i < Ci_runningSampleCount; i++) {
      sprintf(sz_fileName, "run%d.wav", i+1);
      apr_runningSamples[i] = load_sample(sz_fileName);
   }
   return TRUE;
}

/**
 * Introduction: show the game's opening
 */

int Game_intro() {
   int i_key;
   int i_option;
   int i;
   i_option = 0;

   // Intro animation
   Animation_play("intro.fli", apr_songs[0]);
   // Game splash
   clear(pr_backBuffer);
   draw_sprite(pr_backBuffer, pr_title, Ci_resX / 2 - 250, 70);
   draw_sprite(pr_backBuffer, pr_options, Ci_resX / 2 - 55, 280);
   writeText(400, "(C) 2001 Pixellation Studios");
   clear_keybuf();
   set_mod_volume(180);
   play_mod(apr_songs[1], TRUE);
   do {
      blit(pr_lodeIcon, pr_backBuffer, 0, 0,
            Ci_resX / 2 - 70,
            300 + (33 * i_option),
            Ci_spriteW, Ci_spriteH);
      rectfill(pr_backBuffer,
            Ci_resX / 2 - 70,
            300 + (33 * (i_option == 0?1:0)),
            Ci_resX / 2 - 70 + Ci_spriteW - 1,
            300 + (33 * (i_option == 0?1:0)) + Ci_spriteH - 1,
            0x000000);
      blit(pr_backBuffer, screen, 0, 0, 0, 0, Ci_resX, Ci_resY);
      
      i_key = readkey() >> 8;
      
      if (i_key == KEY_UP || i_key == KEY_DOWN) {
         i_option = (i_option == 0?1:0);
         Sound_playSample(GET_TREASURE, 128);
      }
      
      
   } while (!(i_key == KEY_SPACE || i_key == KEY_ENTER));
   for (i = 180; i > 0; i-=5) {
      set_mod_volume(i);
      if (i_option == 1) {
         rectfill(screen, 0, 0, 640, (180-i) * 2.6, 0x000000);
      }
      rest(16+(16*i_option)); // slower fade for exit
   }
   stop_mod();
   if (i_option == 1)
      return FALSE;
   return TRUE;
}

/**
 * Display "Game Over" sequence.
 */

void Game_gameOver() {
   int i;
   int j;
   
   play_mod(apr_songs[3], FALSE);
   
   for (i = 10; i <= 50; i+=10) {
      for (j = 0; j <= 20; j+=2) {
         blit(pr_arenaBuffer, pr_backBuffer,
              Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              0, 0, 240, 40);
         stretch_sprite(pr_backBuffer, pr_gameOver, 0, j, 240, 40-(j*2));
         blit(pr_backBuffer, screen,
              0, 0, Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              240, 40);
         rest(i/(j+1));
      }
      for (j = 20; j >= 0; j-=2) {
         blit(pr_arenaBuffer, pr_backBuffer,
              Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              0, 0, 240, 40);
         stretch_sprite(pr_backBuffer, pr_gameOverFlip, 0, j, 240, 40-(j*2));
         blit(pr_backBuffer, screen,
              0, 0, Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              240, 40);
         rest(i/(j+1));
      }
      for (j = 0; j <= 20; j+=2) {
         blit(pr_arenaBuffer, pr_backBuffer,
              Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              0, 0, 240, 40);
         stretch_sprite(pr_backBuffer, pr_gameOverFlip, 0, j, 240, 40-(j*2));
         blit(pr_backBuffer, screen,
              0, 0, Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              240, 40);
         rest(i/(j+1));
      }
      for (j = 20; j >= 0; j-=2) {
         blit(pr_arenaBuffer, pr_backBuffer,
              Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              0, 0, 240, 40);
         stretch_sprite(pr_backBuffer, pr_gameOver, 0, j, 240, 40-(j*2));
         blit(pr_backBuffer, screen,
              0, 0, Ci_resX / 2 - 120, Ci_resY / 2 - 20,
              240, 40);
         rest(i/(j+1));
      }
   }
   readkey();
   stop_mod();
}

/**
 * Game outer loop: the loop between rounds of gameplay
 * (display init screen, load stages, call inner loop...).
 */

void Game_loop() {

   while (TRUE) {
      if (Game_intro() == FALSE)
         break;
      Play_init();
      while (i_lives >= 0) {
         Play_loadStage();
         if (Play_loop()) {
            Play_nextStage();
            if (i_stage > Ci_stageCount) {
               // then you beat the game? ;-)
               break;
            }
         } else {
            i_lives--;
         }
      }
      Game_gameOver();
      b_gameOver = TRUE;
   }
}

/**
 * Game finalizations; finalizes the game data and
 * utility libraries.
 */

int Game_done() {
   int i;

   clear_keybuf();

   for (i = 0; i < Ci_songCount; i++)
      destroy_mod(apr_songs[i]);
   for (i = 0; i < Ci_sampleCount; i++)
      destroy_sample(apr_samples[i]);
   for (i = 0; i < Ci_ropeSampleCount; i++)
      destroy_sample(apr_ropeSamples[i]);
   for (i = 0; i < Ci_stairsSampleCount; i++)
      destroy_sample(apr_stairsSamples[i]);
   for (i = 0; i < Ci_runningSampleCount; i++)
      destroy_sample(apr_runningSamples[i]);
   for (i = 0; i < Ci_fallSampleCount; i++)
      destroy_sample(apr_fallSamples[i]);
   for (i = 0; i < Ci_ouchSampleCount; i++)
      destroy_sample(apr_ouchSamples[i]);
   
   destroy_bitmap(pr_title);
#ifdef USE_BACKGROUND
   destroy_bitmap(pr_background);
#endif
   destroy_bitmap(pr_backBuffer);
   
   for (i = 0; i < Ci_figureCount; i++)
      if (apr_figures[i] != NULL)
         destroy_bitmap(apr_figures[i]);

   for (i = 0; i < Ci_brokenFrames; i++)
      if (apr_brokenAnimation[i] != NULL)
         destroy_bitmap(apr_brokenAnimation[i]);

   for (i = 0; i < 8; i++) {
      destroy_bitmap(apr_lodeRunningLeft[i]);
      destroy_bitmap(apr_lodeRunningRight[i]);
      destroy_bitmap(apr_monkRunningLeft[i]);
      destroy_bitmap(apr_monkRunningRight[i]);
   }
   destroy_bitmap(pr_lodeFalling);
   destroy_bitmap(pr_monkFalling);

   Allegro_done();
}

/* ***** Main */

/**
 * Main function; does global initializations and finalizations
 * and calls the game's outer loop. 
 */

int main(int i_argc_, char* asz_argv_) {
   int i;
   char sz_filename[20];

   printf("\nXLode\n");
   printf("(C)2001 Hisham Muhammad\n");
   XLode_debug("Opening stages... ");
   
   for (i = 0; i < Ci_stageCount; i++) {
      sprintf(sz_filename, "stage%d.map", i);
      af_stages[i] = fopen(sz_filename, "r");
      ab_cachedStage[i] = FALSE;
   }

   XLode_debug("done.\n");
   XLode_debug("Initializing game... ");

   if (! Game_init() ) {
      printf("Failed initializing.");
      return 1;
   }
   
   XLode_debug("done.\n");
   XLode_debug("Starting game loop.\n");
   
   Game_loop();
   
   XLode_debug("Finished game loop.\n");
   XLode_debug("Closing game... ");

   Game_done();
   
   XLode_debug("done.\n");

   return 0;

}
END_OF_MAIN();
